unit ICQSock;
{(C) Alex Demchenko}
{ Modified by NighTrader 02-04-2003 }

//{$DEFINE DEBUG} {Show debug errors}
//{$DEFINE PARSE} {Dump packets into file}

interface
uses
{$IFDEF DEBUGDC}
  dbugintf, Dialogs,
{$ENDIF}
  WSocket, windows, icqWorks, Classes, winsock;

const
  CNetPktLen = $FFFF; {Size of network packet}
  HTTP_RCV_BUF_SIZE    = 8193;  

type
  PNetPacket = ^TNetPacket;
  TNetPacket = record
    Buf: array[0..CNetPktLen - 1] of Byte;
    BufLen: Word;
    Offset: Word;
    Next: PNetPacket;
  end;

  {Thread-safe implementation of software buffer}
  TNetBuffer = class(TObject)
  private
    FPkt: PNetPacket;
    Shared: Integer;
    CS: TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Enter;
    procedure Leave;
    procedure Clear;
    procedure AddPacket(Buffer: Pointer; BufLen: LongWord);
    procedure DelPacket;
    function GetPacket(Buffer: Pointer): LongWord;
    function SkipData(Len: Word): Boolean;
    procedure AddStr(const Value: String);
    function GetStr: String;
    function GetLength: LongWord;
  end;

  {Thread-safe implementation of Berkeley sockets}
  TOnError = procedure(Sender: TObject; ErrorType: TErrorType; ErrorMsg: String; ErrorAddType: integer) of object;
  TOnRecveive = procedure(Sender: TObject; Buffer: Pointer; BufLen: LongWord) of object;

  {    socks 4  5}
  { +       https }
  TEventSocket = class(TObject)
  private
    FControlSocket: TWSocket;
    FReceiveBuffer        : array [0..HTTP_RCV_BUF_SIZE - 1] of char;
    FProxyType: TProxyType; //     TProxyType = (P_NONE, P_SOCKS4, P_SOCKS5, P_HTTPS);
    FICQHost: string;   //    
    FICQPort: string;   //    
    FProxHost: string;  //  
    FProxPort: string;  //  
    FProxyUser: string; //  
    FProxyUserPass: String; //  
    FProxyAuthorize: boolean; //    

    FProxyReady: boolean;
    FOnProxyError        : TSocksErrorEvent;  //    
    FOnProxyConnected    : TSessionConnected; //        
    FOnSessionClosed     : TSessionClosed; //         

    FOnRecieve           : TOnRecveive; //   
    procedure onConnected (Sender: TObject; ErrCode: Word); //     
    procedure onDataAvailable (Sender: TObject; ErrCode: Word); //    
  public
    constructor Create;
    destructor destroy;override;
    procedure Connect;
    procedure Disconnect;
    property ControlSocket : TWSocket read FControlSocket;
    property ProxyType : TProxyType read FProxyType write FProxyType;
    property ICQHost : string read FICQHost write FICQHost;
    property ICQPort : string read FICQPort write FICQPort;
    property ProxyHost : string read FProxHost write FProxHost;
    property ProxyPort : string read FProxPort write FProxPort;
    property ProxyUser : string read FProxyUser write FProxyUser;
    property ProxyUserPass : string read FProxyUserPass write FProxyUserPass;
    property ProxyAuthorize : boolean read FProxyAuthorize write FProxyAuthorize;

    property OnProxyError : TSocksErrorEvent read FOnProxyError write FOnProxyError;
    property OnProxyConnected : TSessionConnected read FOnProxyConnected write FOnProxyConnected;
    property OnSessionClosed : TSessionClosed read FOnSessionClosed write FOnSessionClosed;

    property OnRecieve : TOnRecveive read FOnRecieve write FOnRecieve;
  end;

  TOnAdvPktParse = procedure(Sender: TObject; Buffer: Pointer; BufLen: LongWord; Incoming: Boolean) of object;
  TOnRecv = procedure(Sender: TObject; Socket: TSocket; Buffer: Pointer; BufLen: LongWord) of object;

function GetHTTPStatus(List: TStringList): String;

implementation

uses sysutils;

const
  b64alphabet: PChar = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';

function EncodeBase64(Value: String): String;
const
  pad: PChar = '====';

  function EncodeChunk(const Chunk: String): String;
  var
    W: LongWord;
    i, n: Byte;
  begin
    n := Length(Chunk); W := 0;
    for i := 0 to n - 1 do
      W := W + Ord(Chunk[i + 1]) shl ((2 - i) * 8);
    Result := b64alphabet[(W shr 18) and $3f] +
              b64alphabet[(W shr 12) and $3f] +
              b64alphabet[(W shr 06) and $3f] +
              b64alphabet[(W shr 00) and $3f];
    if n <> 3 then
      Result := Copy(Result, 0, n + 1) + Copy(pad, 0, 3 - n);   //add padding when out len isn't 24 bits
  end;

begin
  Result := '';
  while Length(Value) > 0 do
  begin
    Result := Result + EncodeChunk(Copy(Value, 0, 3));
    Delete(Value, 1, 3);
  end;
end;

function DecodeBase64(Value: String): String;
  function DecodeChunk(const Chunk: String): String;
  var
    W: LongWord;
    i: Byte;
  begin
    W := 0; Result := '';
    for i := 1 to 4 do
      if Pos(Chunk[i], b64alphabet) <> 0 then
        W := W + Word((Pos(Chunk[i], b64alphabet) - 1)) shl ((4 - i) * 6);
    for i := 1 to 3 do
      Result := Result + Chr(W shr ((3 - i) * 8) and $ff);
  end;
begin
  Result := '';
  if Length(Value) mod 4 <> 0 then Exit;
  while Length(Value) > 0 do
  begin
    Result := Result + DecodeChunk(Copy(Value, 0, 4));
    Delete(Value, 1, 4);
  end;
end;

procedure ShowMessage(const Msg: String);
begin
  MessageBox(0, PChar(Msg), 'Message', 0);
end;

function GetHTTPStatus(List: TStringList): String;
var
  i, c: Word;
  S: String;
begin
  Result := '';
  if List.Count < 1 then Exit;
  S := List.Strings[0]; c := 0;
  for i := 1 to Length(S) do
    if c = 1 then
      Result := Result + S[i]
    else
      if S[i] = ' ' then Inc(c);
end;

constructor TNetBuffer.Create;
begin
  inherited Create;
  FPkt := nil;
  Shared := 0;
end;

destructor TNetBuffer.Destroy;
begin
  Clear;
  if Shared = 1 then DeleteCriticalSection(CS);
  inherited;
end;

{Swap pointers}
procedure XChg(var Critical, Normal); assembler;
asm
  mov  ecx, [edx]
  xchg [eax], ecx
  mov  [edx], ecx
end;

procedure TNetBuffer.Enter; {Synchronization - enter critical section}
var
  j: Integer;
begin
  j := 1; XChg(Shared, j); if j = 0 then InitializeCriticalSection(CS);
  EnterCriticalSection(CS);
end;

procedure TNetBuffer.Leave; {Synchronization - leave critical section}
begin
  LeaveCriticalSection(CS);
end;

procedure TNetBuffer.Clear;
var
  p: Pointer;
begin
  while FPkt <> nil do begin
    p := FPkt^.Next;
    FreeMem(FPkt);
    FPkt := p;
  end;
end;

procedure TNetBuffer.AddPacket(Buffer: Pointer; BufLen: LongWord);
var
  p: PNetPacket;
begin
  if BufLen > CNetPktLen then BufLen := CNetPktLen;
  if FPkt = nil then begin
    GetMem(FPkt, SizeOf(TNetPacket));
    p := FPkt;
  end else begin
    p := FPkt;
    while p <> nil do begin
      if p^.Next = nil then Break;
      p := p^.Next;
    end;
    GetMem(p^.Next, SizeOf(TNetPacket));
    p := p^.Next;
  end;
  p^.BufLen := BufLen;
  p^.Offset := 0;
  p^.Next := nil;
  Move(Buffer^, p^.Buf, BufLen);
end;

procedure TNetBuffer.DelPacket;
var
  p: PNetPacket;
begin
  if (FPkt = nil) then Exit;
  if FPkt^.Next <> nil then
  begin
    p := FPkt^.Next;
    FreeMem(FPkt);
    FPkt := p;
  end else
  begin
    FreeMem(FPkt);
    FPkt := nil;
  end;
end;

function TNetBuffer.GetPacket(Buffer: Pointer): LongWord;
begin
  if (FPkt = nil) or (FPkt^.Offset >= FPkt^.BufLen) then
    Result := 0
  else begin
    Move(Ptr(LongWord(@FPkt^.Buf) + FPkt^.Offset)^, Buffer^, FPkt^.BufLen - FPkt^.Offset);
    Result := FPkt^.BufLen - FPkt^.Offset;
  end;
end;

function TNetBuffer.SkipData(Len: Word): Boolean;
begin
  if FPkt = nil then
    Result := True
  else begin
    Inc(FPkt^.Offset, Len);
    Result := FPkt^.Offset >= FPkt^.BufLen;
  end;
end;

procedure TNetBuffer.AddStr(const Value: String);
begin
  AddPacket(@Value[1], Length(Value));
end;

function TNetBuffer.GetStr: String;
var
  p: array[0..CNetPktLen] of Char;
begin
  p[GetPacket(@p)] := #0;
  Result := PChar(@p);
end;

function TNetBuffer.GetLength: LongWord;
var
  p: PNetPacket;
begin
  Result := 0;
  p := FPkt;
  while p <> nil do begin
    Inc(Result, p^.BufLen);
    p := p^.Next;
  end;
end;


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

{ TCustromSocket }

{** CONSTRUCTOR **}



///-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

{ TEventSocket }

procedure TEventSocket.Connect;
begin
  if FProxyType = p_none then //   
  begin
    FControlSocket.Addr := FICQHost;
    FControlSocket.Port := FICQPort;
    FProxyReady := true;
  end
  else //  
  begin
    //        
    FProxyReady := false;
    if FProxyType <> P_HTTPS then // socks 
    begin
      FControlSocket.Addr := FICQHost;
      FControlSocket.Port := FICQPort;
      FControlSocket.SocksServer := FProxHost;
      FControlSocket.SocksPort := FProxPort;
      FControlSocket.SocksAuthentication := socksNoAuthentication;
      if FProxyAuthorize then
      begin
        FControlSocket.SocksPassword := FProxyUserPass;
        FControlSocket.SocksUsercode := FProxyUser;
        FControlSocket.SocksAuthentication := socksAuthenticateUsercode;
      end;
      //    4  5
      if FProxyType = P_SOCKS4 then FControlSocket.SocksLevel := '4' else FControlSocket.SocksLevel := '5';
    end
    else // https 
    begin
      //    
      FControlSocket.Addr := FProxHost;
      FControlSocket.Port := FProxPort;
      //   
      FControlSocket.SocksServer := '';
      FControlSocket.SocksPort := '';
      FControlSocket.SocksAuthentication := socksNoAuthentication;
      FControlSocket.SocksPassword := '';
      FControlSocket.SocksUsercode := '';
      //
    end;
  end;
  FControlSocket.OnSessionClosed := FOnSessionClosed;
  FControlSocket.Connect;
end;

constructor TEventSocket.Create;
begin
  inherited;
  FOnProxyError := nil;
  FOnProxyConnected := nil;
  FOnRecieve := nil;
  FProxyType := P_NONE;
  FICQHost := 'login.icq.com';
  FICQPort := '5190';
  FProxHost := '';
  FProxPort := '';
  FProxyUser := '';
  FProxyUserPass := '';
  FProxyAuthorize := false;
  FProxyReady := false;
  FControlSocket := TWSocket.Create(nil);
  FControlSocket.Proto := 'tcp';
  FControlSocket.OnSessionConnected := onConnected;
  FControlSocket.OnDataAvailable := onDataAvailable;  
end;

destructor TEventSocket.destroy;
begin
  FreeAndNil(FControlSocket);
  inherited;
end;

procedure TEventSocket.Disconnect;
begin
  FProxyReady := false;
  FControlSocket.Close;
end;


procedure TEventSocket.onConnected(Sender: TObject; ErrCode: Word);
var
  ssend: string;
begin
  ssend := '';
  if ErrCode <> 0 then
  begin
    if assigned(FOnProxyError) then FOnProxyError(Sender, ErrCode, 'error connect host down');
  end
  else
  begin //   
    if FProxyType = P_HTTPS then
    begin  //  https     
      if not (FProxyReady) then
      begin
        ssend := 'CONNECT ' + FICQHost + ':' + FICQPort + ' HTTP/1.0' + #13#10 + 'User-Agent: Mozilla/4.08 [en] (WinNT; U)'+ #13#10;
        if FProxyAuthorize then
        ssend := ssend + 'Authorization: Basic ' + EncodeBase64(FProxyUser + ':' + FProxyUserPass) + #13#10 +
                'Proxy-Authorization: Basic ' + EncodeBase64(FProxyUser + ':' + FProxyUserPass) + #13#10;
        ssend := ssend + #13#10;
        FControlSocket.SendStr(ssend);
      end;
    end
    else
    begin //        ( https)         
      FProxyReady := true;
      if assigned(FOnProxyConnected) then FOnProxyConnected(Sender, 0);
    end;
  end;
end;

procedure TEventSocket.onDataAvailable(Sender: TObject; ErrCode: Word);
var
    Len : Integer;
    I   : Integer;
    VoidBuf : String;
begin
  if ErrCode <> 0 then
  begin
    if assigned(FOnProxyError) then FOnProxyError(Sender, ErrCode, 'error rcv data');
  end
  else
  begin
    i := 0;
    Len := TWSocket(Sender).Receive(@FReceiveBuffer[0], TWSocket(Sender).RcvdCount);
    if Len <= 0 then exit;
    if not FProxyReady then
    begin
      //   -       
      if (
      (StrLIComp(FReceiveBuffer, 'HTTP/1.0 200', 12) = 0) or
      (StrLIComp(FReceiveBuffer, 'HTTP/1.1 200', 12) = 0) or
      (StrLIComp(FReceiveBuffer, 'HTTP/1.0  200', 13) = 0) or // M$ Proxy Server 2.0
      (StrLIComp(FReceiveBuffer, 'HTTP/1.1  200', 13) = 0)    // M$ Proxy Server 2.0 not tested ??
      ) and not
      ((StrLIComp(FReceiveBuffer, 'HTTP/1.1 200 OK', 15) = 0) or
      (StrLIComp(FReceiveBuffer, 'HTTP/1.0 200 OK', 15) = 0)) then
      begin
        I := pos(#13#10#13#10, FReceiveBuffer) + 3;
        setlength(VoidBuf, Len - I);
        Move(FReceiveBuffer[I], VoidBuf[1], Len - I);
        Move(VoidBuf[1], FReceiveBuffer[0], Len - I);
        FProxyReady := true;
        if assigned(FOnProxyConnected) then FOnProxyConnected(Sender, 0);        
      end
      else
      begin
        disconnect;
        if assigned(FOnProxyError) then FOnProxyError(Sender, ErrCode, 'error connect proxy bad');
      end;
    end;
    if (Len-i) > 0 then
      if Assigned (FOnRecieve) then FOnRecieve(Sender, @FReceiveBuffer, Len-i);
  end;
end;

end.
